// hlavickove subory *******************************************************
// systemove funkcie
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
// GL a GLU kniznica
#include <GL/gl.h>
//#include <GL/glu.h>

// .h subor
#include "glTriangle.h"

// konstruktor *************************************************************
glTriangle::glTriangle()
{
 all_tris_count=0; // celkovy pocet trojuholnikov

 m_first_triangle=NULL; // v zozname nie su ziadne trojuholniky
 m_last_triangle=NULL;

 triangle_array=NULL;
}

// destruktor **************************************************************
glTriangle::~glTriangle()
{
 if (m_first_triangle!=NULL) {
  // zrusenie spojiteho pola trojuholnikov
  if (triangle_array!=NULL) free(triangle_array);

  while (m_first_triangle->next!=NULL) { // zmazanie vsetkych trojuholnikov zo zoznamu
   m_t=m_first_triangle->next;
   delete m_first_triangle;
   m_first_triangle=m_t;
  }
  delete m_first_triangle;               // zmazanie 1. trojuholnika
  m_first_triangle=NULL;
 }
}

// normalizacia normaly ****************************************************
void glTriangle::NormaliseNormal(GLfloat n1,GLfloat n2,GLfloat n3,
                                 GLfloat *a,GLfloat *b,GLfloat *c)
{
 GLfloat d;
 // ziskanie velkosti normaly
 d=sqrt(n1*n1+n2*n2+n3*n3);
 // uprava hodnot normaly
 if (d!=0) { *a=n1/d; *b=n2/d; *c=n3/d; }
}

// vypocet normaly *********************************************************
void glTriangle::CalcNormal(GLfloat p1,GLfloat p2,GLfloat p3,
                            GLfloat q1,GLfloat q2,GLfloat q3,
                            GLfloat r1,GLfloat r2,GLfloat r3,
                            GLfloat *a,GLfloat *b,GLfloat *c)
{
 GLfloat v11,v12,v13; // 1. vektor
 GLfloat v21,v22,v23; // 2. vektor
 GLfloat n1,n2,n3,nn1,nn2,nn3;

 v11=q1-p1; v12=q2-p2; v13=q3-p3; // vypocet hodnot vektorov
 v21=r1-p1; v22=r2-p2; v23=r3-p3;

 n1=v12*v23-v13*v22;              // vektorovy sucin vektorov
 n2=v13*v21-v11*v23;
 n3=v11*v22-v12*v21;

 // normalizacia normaly
 NormaliseNormal(n1,n2,n3,&nn1,&nn2,&nn3);
 // vratenie vysledku
 *a=nn1; *b=nn2; *c=nn3;
}

// vypocet normal stran pre vypocet kolizie ********************************
void glTriangle::CalcCollisionNormals(void)
{
 GLfloat v11,v12,v13; // 1. vektor
 GLfloat v21,v22,v23; // 2. vektor
 GLfloat n1,n2,n3,nn1,nn2,nn3;

 v21=m_last_triangle->tr_nx; // 2. vektorom pre sucin je vzdy normala trojuholnika
 v22=m_last_triangle->tr_ny;
 v23=m_last_triangle->tr_nz;

 // normala pre 1. stranu trojuholnika
 v11=m_last_triangle->x2-m_last_triangle->x1; // vypocet hodnot 1. vektora
 v12=m_last_triangle->y2-m_last_triangle->y1;
 v13=m_last_triangle->z2-m_last_triangle->z1;

 n1=v12*v23-v13*v22; // vektorovy sucin vektorov
 n2=v13*v21-v11*v23;
 n3=v11*v22-v12*v21;

 // normalizacia normaly
 NormaliseNormal(n1,n2,n3,&nn1,&nn2,&nn3);
 // vratenie vysledku
 m_last_triangle->tr_cn1=nn1;
 m_last_triangle->tr_cn2=nn2;
 m_last_triangle->tr_cn3=nn3;
 // podla rovnice roviny ax+by+cz+d=0 => d=-(ax+by+cz)
 m_last_triangle->tr_cn4=-(nn1*m_last_triangle->x1+
                           nn2*m_last_triangle->y1+
                           nn3*m_last_triangle->z1);

 // normala pre 2. stranu trojuholnika
 v11=m_last_triangle->x3-m_last_triangle->x2; // vypocet hodnot 1. vektora
 v12=m_last_triangle->y3-m_last_triangle->y2;
 v13=m_last_triangle->z3-m_last_triangle->z2;

 n1=v12*v23-v13*v22; // vektorovy sucin vektorov
 n2=v13*v21-v11*v23;
 n3=v11*v22-v12*v21;

 // normalizacia normaly
 NormaliseNormal(n1,n2,n3,&nn1,&nn2,&nn3);
 // vratenie vysledku
 m_last_triangle->tr_cn5=nn1;
 m_last_triangle->tr_cn6=nn2;
 m_last_triangle->tr_cn7=nn3;
 // podla rovnice roviny ax+by+cz+d=0 => d=-(ax+by+cz)
 m_last_triangle->tr_cn8=-(nn1*m_last_triangle->x2+
                           nn2*m_last_triangle->y2+
                           nn3*m_last_triangle->z2);

 // normala pre 3. stranu trojuholnika
 v11=m_last_triangle->x1-m_last_triangle->x3; // vypocet hodnot 1. vektora
 v12=m_last_triangle->y1-m_last_triangle->y3;
 v13=m_last_triangle->z1-m_last_triangle->z3;

 n1=v12*v23-v13*v22; // vektorovy sucin vektorov
 n2=v13*v21-v11*v23;
 n3=v11*v22-v12*v21;

 // normalizacia normaly
 NormaliseNormal(n1,n2,n3,&nn1,&nn2,&nn3);
 // vratenie vysledku
 m_last_triangle->tr_cn9=nn1;
 m_last_triangle->tr_cn10=nn2;
 m_last_triangle->tr_cn11=nn3;
 // podla rovnice roviny ax+by+cz+d=0 => d=-(ax+by+cz)
 m_last_triangle->tr_cn12=-(nn1*m_last_triangle->x3+
                            nn2*m_last_triangle->y3+
                            nn3*m_last_triangle->z3);
}

// vyhladenie normal, q-pocet troj. v smere x, w-pocet troj. v smere y *****
int glTriangle::SmoothNormals(GLuint q,GLuint w)
{
 GLuint a;
 GLuint a1,a2,a3,a4,a5,a6;

 // INFO: pocet trojuholnikov v smere x a y
 //fprintf(stderr,"x tris_count: %u, y tris_count: %u\n",q,w);

 // docasne pole pre plosne ulozenie trojuholnikov
 GLfloat *m_data;
 m_data=(GLfloat *) malloc(q*w*9*sizeof(GLfloat));
 if (m_data==NULL) {
  fprintf(stderr,"Error allocating memory for m_data!\n"); return(1);
 }
 // ulozenie normal trojuholnikov do plosneho pola
 a=0;
 m_t=m_first_triangle;
 while (m_t!=NULL) {
  *(m_data+a)  =m_t->xn1; *(m_data+a+1)=m_t->yn1; *(m_data+a+2)=m_t->zn1;
  *(m_data+a+3)=m_t->xn2; *(m_data+a+4)=m_t->yn2; *(m_data+a+5)=m_t->zn2;
  *(m_data+a+6)=m_t->xn3; *(m_data+a+7)=m_t->yn3; *(m_data+a+8)=m_t->zn3;
  a+=9;
  m_t=m_t->next;
 }
 // uprava normal - spriemerovanie 6-tich susediacich normal
 for (a=12;a<q*(w-2)*9;a+=18) {
  a1=a;          a2=a1+12; a3=a2+9;
  a4=a1+(q*9-9); a5=a4+6;  a6=a5+9;
  // uprava hodnoty 1. normaly
  *(m_data+a1)  =(*(m_data+a1)+*(m_data+a2)+*(m_data+a3)+
                  *(m_data+a4)+*(m_data+a5)+*(m_data+a6))/6.0f;
  *(m_data+a1+1)=(*(m_data+a1+1)+*(m_data+a2+1)+*(m_data+a3+1)+
                  *(m_data+a4+1)+*(m_data+a5+1)+*(m_data+a6+1))/6.0f;
  *(m_data+a1+2)=(*(m_data+a1+2)+*(m_data+a2+2)+*(m_data+a3+2)+
                  *(m_data+a4+2)+*(m_data+a5+2)+*(m_data+a6+2))/6.0f;
  // vsetky dalsie normaly budu rovnake ako 1. normala
  *(m_data+a2)  =*(m_data+a1);
  *(m_data+a2+1)=*(m_data+a1+1);
  *(m_data+a2+2)=*(m_data+a1+2);

  *(m_data+a3)  =*(m_data+a1);
  *(m_data+a3+1)=*(m_data+a1+1);
  *(m_data+a3+2)=*(m_data+a1+2);

  *(m_data+a4)  =*(m_data+a1);
  *(m_data+a4+1)=*(m_data+a1+1);
  *(m_data+a4+2)=*(m_data+a1+2);

  *(m_data+a5)  =*(m_data+a1);
  *(m_data+a5+1)=*(m_data+a1+1);
  *(m_data+a5+2)=*(m_data+a1+2);

  *(m_data+a6)  =*(m_data+a1);
  *(m_data+a6+1)=*(m_data+a1+1);
  *(m_data+a6+2)=*(m_data+a1+2);
 }
 // vyber normal trojuholnikov z plosneho pola
 a=0;
 m_t=m_first_triangle;
 while (m_t!=NULL) {
  m_t->xn1=*(m_data+a);   m_t->yn1=*(m_data+a+1); m_t->zn1=*(m_data+a+2);
  m_t->xn2=*(m_data+a+3); m_t->yn2=*(m_data+a+4); m_t->zn2=*(m_data+a+5);
  m_t->xn3=*(m_data+a+6); m_t->yn3=*(m_data+a+7); m_t->zn3=*(m_data+a+8);
  a+=9;
  m_t=m_t->next;
 }
 // uvolnenie alokovaniej pamate
 if (m_data!=NULL) free(m_data);

 // INFO: alokovane miesto a obsadene miesto
 //fprintf(stderr,"data=%u, a=%u\n",q*w*9,a);

 return(0);
}

// pridanie trojuholnika do zoznamu ****************************************
int glTriangle::triangle_add(GLfloat a1,GLfloat a2,GLfloat a3,
                             GLfloat b1,GLfloat b2,GLfloat b3,
                             GLfloat c1,GLfloat c2,GLfloat c3,
                             GLfloat d1,GLfloat d2,
                             GLfloat e1,GLfloat e2,
                             GLfloat f1,GLfloat f2)
{
 if (m_first_triangle==NULL) { // vytvorenie 1. trojuholnika
  m_first_triangle=new _TRIANGLE;
  if (m_first_triangle==NULL) {
   fprintf(stderr,"Error allocationg memory for triangle!\n");
   return(1);
  }
  m_last_triangle=m_first_triangle; // na zaciatku m_last_triangle=m_first_triangle
 }
 else {                        // vytvorenie dalsieho trojuholnika
  m_last_triangle->next=new _TRIANGLE;
  if (m_last_triangle->next==NULL) {
   fprintf(stderr,"Error allocationg memory for triangle!\n");
   return(1);
  }
  m_last_triangle=m_last_triangle->next;
 }

 m_last_triangle->id=0;
 m_last_triangle->x1=a1; m_last_triangle->y1=a2; m_last_triangle->z1=a3;
 m_last_triangle->x2=b1; m_last_triangle->y2=b2; m_last_triangle->z2=b3;
 m_last_triangle->x3=c1; m_last_triangle->y3=c2; m_last_triangle->z3=c3;
 CalcNormal(a1,a2,a3,
            b1,b2,b3,
            c1,c2,c3,
            &xx,&yy,&zz);
 m_last_triangle->tr_nx=xx; m_last_triangle->tr_ny=yy; m_last_triangle->tr_nz=zz;
 m_last_triangle->xn1=xx; m_last_triangle->yn1=yy; m_last_triangle->zn1=zz;
 m_last_triangle->xn2=xx; m_last_triangle->yn2=yy; m_last_triangle->zn2=zz;
 m_last_triangle->xn3=xx; m_last_triangle->yn3=yy; m_last_triangle->zn3=zz;
 m_last_triangle->xt1=d1; m_last_triangle->yt1=d2;
 m_last_triangle->xt2=e1; m_last_triangle->yt2=e2;
 m_last_triangle->xt3=f1; m_last_triangle->yt3=f2;
 m_last_triangle->next=NULL;

 CalcCollisionNormals(); // vypocet normal stran trojuholnika potrebnych pre neskorsi
                         // vypocet kolizie s tymto trojuholnikom

 all_tris_count++;

 return(0);
}

// ziskanie odkazu na zoznam trojuholnikov *********************************
_TRIANGLE *glTriangle::GetTriangles(void)
{
 return m_first_triangle;
}

// vykreslenie normal ******************************************************
void glTriangle::DrawNormals()
{
 if (m_first_triangle!=NULL) {
  glPushAttrib(GL_LIGHTING_BIT);
  glDisable(GL_LIGHTING);
  glBegin(GL_LINES);              // ciary
   m_t=m_first_triangle;
   while (m_t!=NULL) {
    // vykreslia sa normaly v kazdom vrchole
    glColor4f(1.0f,0.0f,0.0f,1.0f); // cervena farba
    glVertex3f(m_t->x1,m_t->y1,m_t->z1);
    glVertex3f(m_t->x1+m_t->xn1*20.0f,m_t->y1+m_t->yn1*20.0f,m_t->z1+m_t->zn1*20.0f);
    glVertex3f(m_t->x2,m_t->y2,m_t->z2);
    glVertex3f(m_t->x2+m_t->xn2*20.0f,m_t->y2+m_t->yn2*20.0f,m_t->z2+m_t->zn2*20.0f);
    glVertex3f(m_t->x3,m_t->y3,m_t->z3);
    glVertex3f(m_t->x3+m_t->xn3*20.0f,m_t->y3+m_t->yn3*20.0f,m_t->z3+m_t->zn3*20.0f);

    // vykreslsa sa normaly stran pre vypocet kolizie
    glColor4f(0.0f,0.0f,1.0f,1.0f); // modra farba
    glVertex3f(m_t->x1,m_t->y1,m_t->z1);
    glVertex3f(m_t->x1+m_t->tr_cn1*20.0f,m_t->y1+m_t->tr_cn2*20.0f,m_t->z1+m_t->tr_cn3*20.0f);
    glVertex3f(m_t->x2,m_t->y2,m_t->z2);
    glVertex3f(m_t->x2+m_t->tr_cn4*20.0f,m_t->y2+m_t->tr_cn5*20.0f,m_t->z2+m_t->tr_cn6*20.0f);
    glVertex3f(m_t->x3,m_t->y3,m_t->z3);
    glVertex3f(m_t->x3+m_t->tr_cn7*20.0f,m_t->y3+m_t->tr_cn8*20.0f,m_t->z3+m_t->tr_cn9*20.0f);

    m_t=m_t->next;
   }
  glEnd();
  glPopAttrib();
 }
}

// vytvorenie spojiteho pola trojuholnikov zo zoznamu **********************
int glTriangle::create_triangle_array(void)
{
 if (m_first_triangle!=NULL) {
  // spojite pole bude obsahovat suradnice trojuholnika, normal, koordinatov textur,
  // normal jednotlivych stran pre kolizie + drawbit
  triangle_array=(GLfloat *) malloc(all_tris_count*(3*3)*(4*3)*(6+1)*(3*4)*sizeof(GLfloat));
  if (triangle_array==NULL) {
   fprintf(stderr,"Error allocating memory for triangle_array!\n"); return(1);
  }

  // id bude naplnene adresou trojuholnika v spojitom poli
  a=0;
  m_t=m_first_triangle;
  while (m_t!=NULL) {
   m_t->id=a;
   // suradnice trojuholnika
   *(triangle_array+a  )=m_t->x1; *(triangle_array+a+1)=m_t->y1; *(triangle_array+a+2)=m_t->z1;
   *(triangle_array+a+3)=m_t->x2; *(triangle_array+a+4)=m_t->y2; *(triangle_array+a+5)=m_t->z2;
   *(triangle_array+a+6)=m_t->x3; *(triangle_array+a+7)=m_t->y3; *(triangle_array+a+8)=m_t->z3;
   // hodnoty zakladnej normaly
   *(triangle_array+a+9)=m_t->tr_nx; *(triangle_array+a+10)=m_t->tr_ny; *(triangle_array+a+11)=m_t->tr_nz;
   // hodnoty vyhladenych normal
   *(triangle_array+a+12)=m_t->xn1; *(triangle_array+a+13)=m_t->yn1; *(triangle_array+a+14)=m_t->zn1;
   *(triangle_array+a+15)=m_t->xn2; *(triangle_array+a+16)=m_t->yn2; *(triangle_array+a+17)=m_t->zn2;
   *(triangle_array+a+18)=m_t->xn3; *(triangle_array+a+19)=m_t->yn3; *(triangle_array+a+20)=m_t->zn3;
   // koordinaty textur
   *(triangle_array+a+21)=m_t->xt1; *(triangle_array+a+22)=m_t->yt1;
   *(triangle_array+a+23)=m_t->xt2; *(triangle_array+a+24)=m_t->yt2;
   *(triangle_array+a+25)=m_t->xt3; *(triangle_array+a+26)=m_t->yt3;
   // hodnoty normal stran pre vypocet kolizie
   *(triangle_array+a+27)=m_t->tr_cn1; *(triangle_array+a+28)=m_t->tr_cn2; *(triangle_array+a+29)=m_t->tr_cn3; *(triangle_array+a+30)=m_t->tr_cn4;
   *(triangle_array+a+31)=m_t->tr_cn5; *(triangle_array+a+32)=m_t->tr_cn6; *(triangle_array+a+33)=m_t->tr_cn7; *(triangle_array+a+34)=m_t->tr_cn8;
   *(triangle_array+a+35)=m_t->tr_cn9; *(triangle_array+a+36)=m_t->tr_cn10; *(triangle_array+a+37)=m_t->tr_cn11; *(triangle_array+a+38)=m_t->tr_cn12;
   // drawbit
   *(triangle_array+a+39)=0.0f;

   a+=40;
   m_t=m_t->next;
  }
 }

 return(0);
}

// ziskanie odkazu na spojite pole trojuholnikov ***************************
GLfloat *glTriangle::GetTriangleArray(void)
{
 return triangle_array;
}
